# Helper functions to rewrite observations
import os
import re

def format_at(selector):
    """
    Return a formatter for the nodes denoted by selector
    :param selector:
    :return:
    """
    pass

class Formatter:
    def __init__(selector):
        pass


class Formatter:
    """
    Class the holds the formatters
    """

    def split(self, sep='\n'):
        """
        Split the string into lines
        """
        pass

    def replace_str(self, from_str, to_str):
        """
        Replace a string with another string
        """
        return Formatter()

class Rewritter:
    """
    Class that implements a variety of rewritters.
    """
    def __init__(self):
        pass



